/*
// (c) jschulen-at-gmx.de 2002, free for use, comments are welcome
// compile: gcc ... -lncurses  # see NCURSES-HowTo
//
// raw-mode and speaker sound only works on linux console
// if you leave in raw-mode,
//   press 2+F12+' (german-kbd) or 2+F12+= (us-kbd) for ^CX<ENTER>
//   good idea is: alias X="kbd_mode -a"
//
// Use Morse code as input. See cw_key.html for details.
//
//  coding: F => ..-. => s0010 => 10010 => 0x12 => 18
*/


/*
 *  use ASCII for more as 6 morse bits 0=dit 1=daw
 *  only a..z,0..9 are original codes 
 *  use free morse codes as modifier
 *    possible modifiers: shift, alt, altgr, cw-command, ESC, control
 *    examples: SHIFT-a=~a=uppercase-A, ^c=CTRL-C
 */
const char morse[]= /* expand*2 for UTF8-16bit? */
  "__etianmsurwdkgohvf~l@pjbxcyzq&^"  /* 1..4 bits */
  "54.3<>,2_-+*#=%16=/\\()[]7{}|8;90"; /* 5 bits */
char *cw1=
"  1\\2      .   -    ..  .-  -.  --     ... ..- .-. .-- -.. -.- --. --- \n"
 "  .  E  .  I   A  .  S   U   R   W   .  H   V   F  +~+  L  +@+  P   J  \n"
 "  -  T  -  N   O  -  D   K   G   O   -  B   X   C   Y   Z   Q  +&+ +^+ \n"
"1\\2 ... ..- .-. .-- -.. -.- --. ---    ... ..- .-. .-- -.. -.- --. ---\n"
 " ..  5   4   .   3   <   >   ,   2  .-  _   -   +   *   #   =   %   1 \n"
" -.  6   =   /   \\   (   )   [   ]  --  7   {   }   |   8   ;   9   0 \n"
 " RET=[SC] DEL=[SB] !=[RV] \"=[RF] $=[RL]   ^M=0x0D=RET  ~a=A ~b=B ~~=~\n"
 " SPC=[RH] ESC=[UY] `=[GH]                  ASCII-7: 0=dit 1=daw\n"
 " &cw-command& &bye; &help;\n";

/*
"      ... ..- .-. .-- -.. -.- --. ---    ... ..- .-. .-- -.. -.- --. ---\n"
"... .                                  - DEL     RET                    \n"
"..- .                                  -             ESC                \n"
".-. . SPC  !   \"   #   $   %   &   '   -  (   )   *   +   ,   -   .   /\n"
"--. .  0   1   2   3   4   5   6   7   -  8   9   :   ;   <   =   >   ? \n"

  "  RR .-..-.  '  WG .----.  $  SX ...-..- (  KN -.--.   )  KK -.--.-
  +  AR .-.-.   ,  GW --..--  -  DU -....-  =  BT -...-   ?  UD ..--..
  _  IQ ..--.-  .  RK .-.-.-  ;  KR -.-.-.  :  OS ---...  /  DN -..-.

 Control_Codes (ToDo): man console_codes
   &D txt morse ...; = define newcode
   &takt dot-time; = set time for dot in ms
 
*/
#ifndef NO_NCURSES
#define HAVE_NCURSES        // check: rpm -q ncurses
#endif
#include <sys/time.h>
#include <sys/ioctl.h>
#include <sys/types.h> // fork()
#include <fcntl.h>     // needed for open()
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <linux/kd.h>
#include <linux/soundcard.h>
#ifdef HAVE_NCURSES
#include <ncurses.h>
#endif
#include <unistd.h>

int nx=80,ny=25;
int imode=3;     /* input-mode, see help */
int omode=0;    /* output mode 0=none, 1=morse, 2=spelling (ToDo) */
char *opipe=NULL; /* output pipe */
int fs=-1;	/* sound-device if positive */
int tt=100;     /* time for a dot in ms */
int sp=2;       /* extra pause between chars of (sp*tt) ms */
int combine=0;  /* combine next character with Ctrl-^, Alt-@, Shift-~ */

#define COMBINE_CTRL  1    /* flags for combine modes */
#define COMBINE_ALT   2
#define COMBINE_SHIFT 4
#define COMBINE_CMD   8

#define TRIGGER_CTRL  '^'  /* chars triggering combine modes */
#define TRIGGER_ALT   '@'
#define TRIGGER_SHIFT '~'
#define TRIGGER_CMD   '&'

void my_wait(int ms){
  struct timeval tv;
  tv.tv_sec=ms/1000;
  tv.tv_usec=1000*(ms%1000);
  select(0,NULL,NULL,NULL,&tv);
}

// output of <SPC>.- in line 2 and sound out
void outsound(char x){
  int i,status; static char buf[4]={'\0','\200','\377','\200'};
  // out sound to make 2 and 3 key mode easier, 1 key mode should be async
  //  close and open cost lot of time, but I do not know,
  //  how to write to /dev/dsp without 2kB-buffering
  // better use nonblocking- or background prozess
#if 0
  printf("%c",x);fflush(stdout); // log this?
#endif
  if (fs>=0) { // out sound
   if (fs>0) { // soundcard 8kHz=0.125ms=1/8ms
    if (x=='.') for(i=0;i<  2*tt;i++) write(fs,buf,4);
    if (x=='-') for(i=0;i<3*2*tt;i++) write(fs,buf,4);
    if (x==',') for(i=0;i<  2   ;i++) write(fs,buf+1,3);
    if (x==' ') for(i=0;i<(sp+1)*8*tt;i++) write(fs,buf+1,1);
    for(i=0;i<  2*tt;i++) write(fs,buf+1,1);
    status = ioctl(fs, SOUND_PCM_SYNC, 0); // takes long time!
    if (status == -1) perror("SOUND_PCM_SYNC ioctl failed");
   } else { // speaker
    if (x=='.' || x=='-'){  // man console_ioctl
      ioctl(1,KIOCSOUND,0x5cf);
      my_wait(tt*((x=='-')?3:1));
      ioctl(1,KIOCSOUND,0);
      my_wait(tt);
    }
    if (x==' ' || x==',') { // ,= short burb as ACK
      my_wait(10);
      ioctl(1,KIOCSOUND,0xb9f);
      my_wait(10);
      ioctl(1,KIOCSOUND,0);
      my_wait(((x==' ')?(sp+1)*tt-20:10));
    }
   }
  }
}

void a2morse(char x){
  int pos,bits;
  if (x==' ') { outsound(' '); return; }
  if ((x^0x20)>='a' && (x^0x20)<='z') { a2morse(TRIGGER_SHIFT);x^=0x20; }
  if ((x^0x60)>='a' && (x^0x60)<='z') { a2morse(TRIGGER_CTRL );x^=0x60; }
  pos=strchr(morse+2,x)-morse;
  if (pos<1 || pos>=1<<6) { pos=0x100 | x; /* unknown ASCII */ }
  for(bits=0;pos>>(bits+1);bits++); // count morse bits
  for(;bits>0;bits--) outsound((1&(pos>>(bits-1)))?'-':'.');
  outsound(' ');
}

// output of speech or morse
// ToDo: read ascXX.[au,wav|mp3] and output, cache files?
//       pipe to flite or other speech-pg if device is not blocked?
void dictate(char *txt){
  static FILE *f1=NULL; int i;
  if (omode==1) {
    for (i=0;txt[i];i++) a2morse(txt[i]);
  }
  if (!opipe) return;
  if (!f1) { f1=popen(opipe,"w"); }
  if (f1) { fprintf(f1,txt); pclose(f1); f1=NULL; }
}

// output of <SPC>.- in line 2 and sound out
void outbit(char x){
#ifdef HAVE_NCURSES
  static int ix=8,iy=11,i;
  ix++; if(ix>=nx) for(i=ix=8;i<nx;i++) mvprintw(iy,i," ");
  mvprintw(iy,ix,"%c",x); refresh();
#endif
  if ((imode&3)>1) outsound(x);
}

// output emulated key to pipe-fd
void outkey(char x, int fd){
  static char cmd[16]; static int cmdlen=0;
#ifdef HAVE_NCURSES
  static int ix=8,iy=12; int i;
  ix++; if (ix>=nx || x=='\n') for(ix=i=8;i<nx;i++) mvprintw(iy,i," "); 
  mvprintw(iy,ix,"%c",x); refresh();
#endif
  // combine codes as seperate function?
  switch (x) {
    case TRIGGER_CTRL : if  (combine & COMBINE_CTRL ) combine=0;
                        else { combine=COMBINE_CTRL ; return; } break;
    case TRIGGER_ALT  : if  (combine & COMBINE_ALT  ) combine=0;
                        else { combine=COMBINE_ALT  ; return; } break;
    case TRIGGER_SHIFT: if  (combine & COMBINE_SHIFT) combine=0;
                        else { combine=COMBINE_SHIFT; return; } break;
    case TRIGGER_CMD  : if (!combine) { combine=COMBINE_CMD; return; } break;
    default: ;
  }
  if (combine & COMBINE_CTRL ) { x^=0x60; }; // ^a=1,^b=2,...
  if (combine & COMBINE_SHIFT) { x^=0x20; }; // ~a=A,~b=B,...
  if (combine & COMBINE_CMD  ) {
    if (cmdlen>15 || x==TRIGGER_CMD || x==';' || x<'0' || x>'z') { // action
      if (!strncmp(cmd,"help",4)) ;
      if (!strncmp(cmd,"sound",5)) ;
      if (!strncmp(cmd,"exit",4)) exit(0);
      if (!strncmp(cmd,"quit",4)) exit(0);
      if (!strncmp(cmd,"bye",3)) exit(0);
      if (!strncmp(cmd,"takt",4)) tt=atoi(cmd+4);
      cmdlen=0; cmd[cmdlen]=0; return;  // reset buf
    } else { // add char to buf
      cmd[cmdlen]=x; cmdlen++; cmd[cmdlen]=0; return;
    }
  }
  // ToDo: how to combine ALT and CMD? (sound on/off etc?)
  if (fd>0) write(fd,&x,1);
  combine&=COMBINE_CMD;
  // fprintf(stderr," combine=%02x x=%02x\n",combine,x);
}

// outchar of <SPC>.- in line 6-ny
void outch(char x){
#ifdef HAVE_NCURSES
  static int ix=0,iy=14; int i,y0=14;
  if (x==8) { mvprintw(iy,ix,"_"); if (ix>0) ix--; refresh(); return; }
  ix=(ix+1)%nx; 
  if((!ix) || x=='\n'){
    iy=(iy-y0+1)%(ny-y0)+y0; ix=0;
    for(i=0;i<nx;i++) mvprintw(iy,i," "); 
  }
  mvprintw(iy,ix,"%c",x); refresh();
#else
  write(1,&x,1);
#endif
}

// outstring in line 6-ny
void outs(char *x){
  int i;
  for (i=0;i<256 && x[i];i++) outch(x[i]);
}

int ms(){
  struct timeval tv; gettimeofday(&tv,NULL); 
  return tv.tv_usec/1000 + 1000*tv.tv_sec;
}

void help(){
  printf(
    "CW_Key 0.2 let you control the computer with only 1-3 keys.\n"
    "You can also use mouse buttons.\n"
    "Therefore the Morse code is used, also known as CW.\n"
    "Options:\n"
    " -1   start in one-key mode (only on tty)\n"
    " -2   start in two-key mode (one key is dit, another daw)\n"
    " -3   start in three-key mode (1st key is dit, 2nd is daw, 3rd is pause)\n"
    " -4   <SPC> by long pause (else <SPC>=[RH]=.-.....)\n"
    " -v   be verbose\n"
    " -t#  set dit-time in ms, default is 100\n"
    " -p#  set extra spacing in dit-times, default is +2\n"
    " -s   use sound (for mode -2 and -3)\n"
    " -sp  use speaker sound\n"
    " -sd  use sound card /dev/dsp 8kHz mono 8bit\n"
    " -O1  output as morse code using speaker or soundcard\n"
    " -o<pipe>  output to a command (speech-synthesizer, cw-encoder)\n"
    );
  exit(0);
}

/* ToDo: write in cw_err.log or err_buf */
#define E1(xx,yy) { perror("perror");fprintf(stderr,"ERROR: L%d "xx"\n",__LINE__,yy);sleep(2); }
#define E0(xx) { perror("perror");fprintf(stderr,"ERROR: L%d "xx"\n",__LINE__);sleep(2); }
#define F0(xx) { E0(xx);exit(1); }

// pipes [0]=readable [1]=writable (0=unused)
int p1[2]={0,0},p2[2]={0,0},p3[2]={0,0};
pid_t child;

#define BSIZE 2
int main(int argn, char *argv[]){ // AE=[AA]=.-.-, OE=[OE]=---.
  int i1,i2,i,j,fd,t0,t1,t2,nds,rc,
      vvv=0,mt=1,space=0,num3=0,n1,n2,n3;
  char buf[BSIZE];  // for child-i/o
  fd_set fileSet;
  fd_set *readFDs = &fileSet;
  struct timeval timeout;
#ifdef HAVE_NCURSES
  WINDOW *w,*w2;
  mmask_t mmask;  // mouse
  MEVENT event;
#endif
  
  for(j=1;j<argn;j++){
    if (argv[j][0]=='-'){
      if (argv[j][1]=='v') vvv++;
      if (argv[j][1]=='1') imode=(imode & ~3) | 1;
      if (argv[j][1]=='2') imode=(imode & ~3) | 2;
      if (argv[j][1]=='3') imode=(imode & ~3) | 3;
      if (argv[j][1]=='4') imode|=4;
      if (argv[j][1]=='h') help();
      if (argv[j][1]=='o') opipe=argv[j]+2;
      if (argv[j][1]=='O') { omode=atoi(argv[j]+2); }
      if (argv[j][1]=='t') { tt=atoi(argv[j]+2); }
      if (argv[j][1]=='p') { sp=atoi(argv[j]+2); }
      if (argv[j][1]=='s') { fs=0; // use sound (speaker if 0)
         if (argv[j][2]==0 || argv[j][1]=='d') {
             fs=open("/dev/dsp",O_WRONLY);
             if(fs<0){ E0("open /dev/dsp");F0("stopped");}
             if (fs) {
               // first: set BufferSize
               // 0xnnnnssss n=max# fragments 2..0x7fff=no_limit
               //            2^sss=buffersize
               i=0x00180004; // buffer must be small to get sound immediately
               if ( ioctl(fs, SNDCTL_DSP_SETFRAGMENT, &i) == -1 ) {
                  E0("ioctl failed"); F0("stopped"); }
               // E1(" ioctl returns: %08x",i);
             }
         }
      }
    } else F0("invalid argument, try option -h for help");
  }

  if (pipe(p1)==-1) E0("pipe");   //  cw_key => p1 => child => p2 => cw_key
  if (pipe(p2)==-1) E0("pipe");   //                   \err => p3 => cw_key
  if (pipe(p3)==-1) E0("pipe");
  child=0;
  child=fork(); if(child==-1) E0("fork");
  if(child){
    close(p1[1]); /* write */
    close(p2[0]); /* read */
    if(close(0)==-1) F0("close0"); if(dup(p1[0])!=0) F0("dup0"); /* read */
    if(close(1)==-1) F0("close1"); if(dup(p2[1])!=1) F0("dup1"); /* write */
    if(close(2)==-1) F0("close2"); if(dup(p3[1])!=2) F0("dup2"); /* werr */
    execlp("sh","sh","-i",NULL);
    F0("exec");
  }
  close(p1[0]); /* read */
  close(p2[1]); /* write */
  close(p3[1]); /* werr */

  if((imode&3)==1){
    fd=open("/proc/self/fd/0",O_RDONLY); 
    if (fd<1) { E0("open /proc/self/fd/0"); }
    if (fd<1)  fd=open("/dev/tty",O_RDONLY); 
    if (fd<1) { E0("open /dev/tty, use mouse!"); }
    if(fd>0) 
      if (ioctl(fd,KDSKBMODE,2)) F0("kbd_mode -s");
  }

  
  // sleep(1); // wait for possible child exit
#ifdef HAVE_NCURSES
  w=initscr();
//  cbreak(); // curses: switch off line buffer
  raw(); // like cbreak, but do not interprete ^S^Q etc.
  noecho();
  timeout(tt/2);
  keypad(w,TRUE); // return one code instead of ESC-sequence (mouse)
  mmask=BUTTON1_RELEASED+BUTTON1_PRESSED
       +BUTTON2_RELEASED+BUTTON2_PRESSED
       +BUTTON3_RELEASED+BUTTON3_PRESSED;
  if ((imode&3)!=1) mmask=BUTTON1_PRESSED+BUTTON2_PRESSED+BUTTON3_PRESSED;
  mmask=mousemask(mmask,NULL); // if 0 not available

  mvprintw(0,0,"cwkey (ESC=quit) t=%4dms spacing=%2ddots sound-fd=%d\n\n%s",
    tt,sp,fs,cw1);
  mvprintw(11,0,"cw-in: ");
  mvprintw(12,0,"key: ");
  refresh();
  w2=newwin(10,78,13,0);wrefresh(w2);scrollok(w2,TRUE);
#else
  printf("cwkey (ESC=quit) t=%4dms spacing=%2ddots sound-fd=%d\n\n%s",
    tt,sp,fs,cw1);
  // setvbuf(stdin,NULL,_IONBF,0);  // does not help! why?
  // -echo: do not show pressed keys
  // -nl: output cr after nl (does not work)
  system("stty raw -echo");  // ioctl() TCGETS TCSETSW
#endif

  n1=n2=n3=0; // count chars and show, for debugging
  t1=t0=ms();
  // morse-input
  for (i1=i2=-1;;){
   // ToDo: use select()
   FD_ZERO(readFDs);
   FD_SET(0, readFDs);           /* stdin */
   FD_SET(p2[0], readFDs);       /* file descriptors */
   FD_SET(p3[0], readFDs);       /* file descriptors */
   nds=p3[0]+1;
   
   timeout.tv_sec = 1; timeout.tv_usec = 0; /* editet by js */
 
   rc = select(nds, readFDs, 0, 0, &timeout);    /* wait for input */
   if (rc < 0) F0("select");

   i2=i1=-1;
   if (FD_ISSET(0, readFDs))
   {
     //       i = read(0, buf, sizeof(buf));
     //       if (i <= 0) F0("read stdin");
     n1++;
#ifdef HAVE_NCURSES
     i2=i1=wgetch(w);
#else
     i2=i1=fgetc(stdin);
#endif
   }

   if (FD_ISSET(p2[0], readFDs))
   {
     i = read(p2[0], buf, sizeof(buf)-1);
     if (i == 0) F0("read pipe2");
     buf[i]=0; n2+=i;
#ifdef HAVE_NCURSES
     wprintw(w2,buf);wrefresh(w2);
#else
     printf(buf);fflush(stdout);
#endif
     dictate(buf);
   } else // stderr after stdout

   if (FD_ISSET(p3[0], readFDs))
   {
     i = read(p3[0], buf, sizeof(buf)-1);
     if (i == 0) F0("read pipe3");
     buf[i]=0; n3+=i;
#ifdef HAVE_NCURSES
     wprintw(w2,buf);wrefresh(w2);
#else
     printf(buf);fflush(stdout);
#endif
     dictate(buf);
   }


   if ((imode&3)!=3) t2=t1=ms(); // KEY_MOUSE=0631
   if ((i1==3 || i1==27) && i1>0) break; // ESC
#ifdef HAVE_NCURSES
   // DEBUG: test curses-beep
   if (i1=='x') { beep(); my_wait(100); }
   if (i1=='y') { beep(); my_wait(100); beep(); my_wait(100); beep(); my_wait(100); }

   if (i1==KEY_MOUSE) if (getmouse(&event) == OK) {
   // daw   = middle button = two   fingers on touchpad of notebooks 
   // space = right  button = three fingers on touchpad of notebooks 
     if (event.bstate&BUTTON1_PRESSED) i1=2; else // dit
     if (event.bstate&BUTTON2_PRESSED) i1=1; else // daw
     if (event.bstate&BUTTON3_PRESSED) i1=3; else // space
     if (event.bstate&BUTTON1_RELEASED) i1=2+128; else
     if (event.bstate&BUTTON2_RELEASED) i1=1+128; else
     if (event.bstate&BUTTON3_RELEASED) i1=3+128; else i1=-1; 
     i2=i1;
   }
#endif
   // translate keys
   if ((imode&3)>1 && i1<128 && i1>0){
     if(strchr(".:aAfFjJ2",i1)) i2=i1=2; // dit
     if(strchr("-_sSgGkK1",i1)) i2=i1=1; // daw
     if(strchr(",;dDhHlL3",i1)) i2=i1=3; // gap
   }
   if (i2>0 && i2<127) {
     if (i2==3) num3++; else num3=0; // count gaps, if num3>1 generate SPACE
     if (num3>1 && i2==3) outkey(' ',p1[1]);
   }
   if ((imode&3)==1) {
     if (i1>0 && i1<128 && fs>=0) ioctl(1,KIOCSOUND,0x5cf);
   }
   if ((imode&3)==3) {
     if (i1<0) continue;    // timeout
     if(t1==t0) t1=t0+tt;
     if (i2>0 && (i2&3)==3){
       t2=t1=t1+(2+sp)*tt; i2=i1=-1; 
     }
   }
   if ((imode&3)>1 && i1>0) { t2=t1+((i2&1)?3   :1)*tt; }
   if(t1-t0>(2+sp)*tt){  // end-of-pause or new-key
     if (mt>1) {
       if(mt>1    && mt<64)   { outkey(morse[mt],p1[1]); } /* 1+6 bits MORSE */
       if(mt>=128 && mt<128+ 63) { outkey(mt& 63,p1[1]); } /* 1+7 bits ASCII */
       if(mt>=256 && mt<256+127) { outkey(mt&127,p1[1]); } /* 1+8 bits ASCII/UTF8 */
       if(mt==1<<7)                   { outkey(8,p1[1]); }
       mt=1;
     }
     if (!space) outbit(' '); space|=1;
     // space between chars is 3, between words is 7
     if (t1-t0>(5+2*sp)*tt && !(space&2)) {
       if (imode&4) outkey(' ',p1[1]);
       outbit(' ');space|=2;
     }
   }
   if (i1<0) continue;
   if ((imode&3)==1) {
     if (i1<16 || i1>64) continue;  // allow only normal scan codes
     // wait for release or time out
#ifdef HAVE_NCURSES
     while((i2=wgetch(w))<128);
#endif
     t2=ms();
     if(fs>=0) ioctl(1,KIOCSOUND,0);
   }
   space=0;
   mt=(mt<<1)+((t2-t1>2*tt)?1:0);  // ToDo: check for OverFlow
#ifdef HAVE_NCURSES
   if (vvv) mvprintw(11, 0,"key %04x %04x  time %4d %4d code=%04x\n",
              i1,i2,t1-t0,t2-t1,mt);
   if (vvv) mvprintw(11,40," keyb %4d p-in: %4d p-err: %4d\n",
              n1,n2,n3);
#endif
   t1=t0=ms();
   outbit( (mt&1) ? '-' : '.' );

   
  }

#ifdef NCURSES
  endwin(); // why console is changed after exit?
#else
#endif
  if((imode&3)==1){
    if (fd>=0) if (ioctl(fd,KDSKBMODE,1)) F0("kbdmode -a");
  }
  if (fs>0) close(fs);
  system("stty sane");
  printf("exit with keycode=%04x\n",i1);
  return(0);
}
